"""
--- Day 15: Timing is Everything ---
https://adventofcode.com/2016/day/15
"""
from collections import deque
from itertools import count

from aocd import data


def parsed(data):
    discs = []
    for line in data.splitlines():
        words = line.split()
        n_pos, pos_t0 = int(words[3]), int(words[-1].rstrip("."))
        disc = deque(range(n_pos))
        disc.rotate(pos_t0)
        discs.append(disc)
    # MAGIC TRICK: if we initially offset each disc by the position it is
    # placed from the top of the stack, then we can assume that the capsule
    # falls through the discs without any delay
    for i, disc in enumerate(discs, 1):
        disc.rotate(i)
    return discs


def find_time(data):
    discs = parsed(data)
    for t in count():
        if not any(disc[0] for disc in discs):
            return t
        for disc in discs:
            disc.rotate(1)


print(find_time(data))
data += "\nBonus disc has 11 positions; at time=0, it is at position 0."
print(find_time(data))



"""
--- Day 5: Sunny with a Chance of Asteroids ---
https://adventofcode.com/2019/day/5
"""
from aoc_wim.aoc2019 import IntComputer
from aocd import data


def compute(data, input_val):
    comp = IntComputer(data, inputs=[input_val])
    comp.run()
    result, *zeros = comp.output
    for zero in zeros:
        assert zero == 0
    return result


if __name__ == "__main__":
    print("part a:", compute(data, input_val=1))
    print("part b:", compute(data, input_val=5))



"""
--- Day 11: Dumbo Octopus ---
https://adventofcode.com/2021/day/11
"""
from aocd import data
from aoc_wim.zgrid import ZGrid


g = ZGrid(data, transform=int)
a = b = 0
while True:
    b += 1
    n = 0
    flash = []
    for z in g:
        if g[z] == 9:
            g[z] = 0
            flash.append(z)
        else:
            g[z] += 1
    while flash:
        z0 = flash.pop()
        n += 1
        for z in g.near(z0, n=8):
            if g.get(z):
                if g[z] == 9:
                    g[z] = 0
                    flash.append(z)
                else:
                    g[z] += 1
    if b <= 100:
        a += n
    if n == len(g):
        break

print("part a:", a)
print("part b:", b)



"""
--- Day 19: Beacon Scanner ---
https://adventofcode.com/2021/day/19
"""
from aocd import data
from parse import parse
import numpy as np
from itertools import combinations
import heapq
import networkx as nx
from scipy.spatial.transform import Rotation as R

Rx = R.from_rotvec([np.pi/2, 0, 0]).as_matrix().astype(int)
Ry = R.from_rotvec([0, np.pi/2, 0]).as_matrix().astype(int)
Rz = R.from_rotvec([0, 0, np.pi/2]).as_matrix().astype(int)
T6 = [np.eye(3, dtype=int)]
for _ in range(3):
    T6.append(np.dot(T6[-1], Rz))
T6.append(Ry)
T6.append(np.linalg.inv(Ry).astype(int))
Ts = []
for T in T6:
    Ts.append(T)
    for _ in range(3):
        Ts.append(np.dot(Ts[-1], Rx))


class Scanner:

    graph = nx.Graph()

    def __init__(self, id, beacons):
        self.id = id
        self.beacons = beacons
        self.transformed = [np.dot(beacons, T) for T in Ts]
        # set of distances between pairs of beacons - this should be invariant under
        # rotations and translations, allows to pair scanners with unknown orientations
        # self.fingerprint = {tuple(abs(x - y)) for x, y in combinations(beacons.T, 2)}
        self.fingerprint = {tuple(sorted(abs(x - y))) for x, y in combinations(beacons, 2)}
        self.transforms = {
            # beacon_id: pose, offset
            self.id: (np.eye(3, dtype=int), np.zeros(3, dtype=int)),
        }

    def __and__(self, other):
        """returns the number of beacons in common (if >= 12, None otherwise)"""
        if not isinstance(other, Scanner):
            return NotImplemented
        for i, other_beacons in enumerate(other.transformed):
            for other_beacon in other_beacons:
                for self_beacon in self.beacons:
                    dt = self_beacon - other_beacon
                    n = len({tuple(b) for b in self.beacons} & {tuple(b) for b in (other_beacons + dt)})
                    assert n >= 1
                    if n >= 12:
                        print(f"scanner {self.id} & {other.id}: {n} common beacons, offset={dt.tolist()}, pose={Ts[i].tolist()}")
                        self.transforms[other.id] = Ts[i], np.dot(Ts[i], -dt)
                        other.transforms[self.id] = np.linalg.inv(Ts[i]).astype(int), dt
                        Scanner.graph.add_edge(self.id, other.id)
                        return n


scanners = {}
for chunk in data.split("\n\n"):
    header, *lines = chunk.splitlines()
    [i] = parse("--- scanner {:d} ---", header).fixed
    numbers = sorted(parse("{:d},{:d},{:d}", line).fixed for line in lines)
    scanner = Scanner(id=i, beacons=np.array(numbers))
    scanners[i] = scanner
    Scanner.graph.add_node(i)

pq = []
for i, (s1, s2) in enumerate(combinations(scanners.values(), 2)):
    # queue pairs of scanners in order of how well their fingerprints matched
    n = len(s1.fingerprint & s2.fingerprint)
    if n >= 12:
        heapq.heappush(pq, (-n, i, s1, s2))
while pq:
    _, _, s1, s2 = heapq.heappop(pq)
    n = s1 & s2
    if n is not None:
        if [*nx.connected_components(Scanner.graph)] == [scanners.keys()]:
            # graph is completely connected: the beacons detected by each scanner may
            # now be transformed into the coordinate system of any other scanner
            break

global_beacons = set()
for i, scanner in scanners.items():
    # transform beacons from this scanner into coordinate system of scanner0
    # note: t0 will also represent this scanner position relative to scanner0
    path = nx.shortest_path(Scanner.graph, scanner.id, 0)
    assert path[0] == i and path[-1] == 0
    beacons = scanner.beacons
    t0 = np.array([0, 0, 0])
    for id0, id1 in zip(path, path[1:]):
        T, dt = scanners[id1].transforms[id0]
        beacons = np.dot(beacons - dt, T)
        t0 = np.dot(t0 - dt, T)
    global_beacons.update({tuple(b) for b in beacons})
    scanner.t0 = t0

print("part a:", len(global_beacons))
print("part b:", max(abs(s0.t0 - s1.t0).sum() for s0, s1 in combinations(scanners.values(), 2)))



"""
--- Day 25: Four-Dimensional Adventure ---
https://adventofcode.com/2018/day/25
"""
from itertools import combinations
import networkx as nx
from aocd import data

nodes = [tuple(int(n) for n in s.split(",")) for s in data.splitlines()]
graph = nx.Graph()
graph.add_nodes_from(nodes)
for node1, node2 in combinations(nodes, 2):
    if sum(abs(x - y) for x, y in zip(node1, node2)) <= 3:
        graph.add_edge(node1, node2)
print(nx.number_connected_components(graph))



"""
--- Day 10: The Stars Align ---
https://adventofcode.com/2018/day/10
"""
import re
import numpy as np
from aocd import data
from aoc_wim.ocr import AOCR
from aoc_wim.zgrid import ZGrid


def varianceish(a):
    return a.std(axis=1)[:2].sum()


numbers = re.findall(r"-?\d+", data)
a = np.array(numbers).astype(int).reshape(-1, 4).T
t = 0
minv = varianceish(a)
while True:
    a[:2] += a[2:]
    v = varianceish(a)
    minv = min(v, minv)
    if v > minv:
        ps = a[:2] - a[2:]  # back up one step
        g = ZGrid({complex(*p): 1 for p in ps.T})
        print("part a:", AOCR[g])
        print("part b:", t)
        break
    t += 1



"""
--- Day 20: A Regular Map ---
https://adventofcode.com/2018/day/20
"""
import networkx as nx
from aocd import data
from aoc_wim.zgrid import manhattan_distance
from aoc_wim.zgrid import ZGrid


def render(g):
    if isinstance(g, str):
        g = graph(g)
    xs = [int(z.real) for z in g.nodes]
    ys = [int(z.imag) for z in g.nodes]
    x0, x1 = min(xs) - 1, max(xs) + 2
    y0, y1 = min(ys) - 1, max(ys) + 2
    for y in range(y0, y1):
        line = ""
        for x in range(x0, x1):
            z = complex(x, y)
            if z == 0:
                line += "X"
            elif z in g.nodes:
                line += "."
            elif (z - 1, z + 1) in g.edges:
                line += "|"
            elif (z - 1j, z + 1j) in g.edges:
                line += "-"
            elif any(zn in g.nodes for zn in ZGrid.near(z, n=8)):
                line += "#"
            else:
                line += " "
        print(line)
    print()


def graph(data, z0=0):
    g = nx.Graph()
    g.add_node(z0)
    tails = [z0]
    stack = []
    for s in data[1:-1]:
        if s in "NSEW":
            dz = 2*getattr(ZGrid, s)
            tails[:] = [z + dz for z in tails]
            for z in tails:
                g.add_edge(z - dz, z)
        elif s == "(":
            stack.append(tails)
            new_tails = tails = tails.copy()
        elif s == "|":
            new_tails.extend(tails)
            tails = stack[-1].copy()
        elif s == ")":
            new_tails.extend(tails)
            stack.pop()
            tails = [*{}.fromkeys(new_tails)]
    return g


g = graph(data)

distances = {}
for z in sorted(g.nodes, key=manhattan_distance, reverse=True):
    if z in distances:
        continue
    path = nx.shortest_path(g, 0, z)
    for length, z in enumerate(path):
        distances[z] = length

print("part a:", max(distances.values()))
print("part b:", sum(d >= 1000 for d in distances.values()))



"""
--- Day 10: Cathode-Ray Tube ---
https://adventofcode.com/2022/day/10
"""
from aocd import data
from aoc_wim.ocr import AOCR


lines = iter(data.splitlines())
cycles = {20 + 40*i for i in range(6)}
W, H = 40, 6
crt = {}
x = 1
a = addx = t = 0
for i in range(W * H):
    if i in cycles:
        a += i * x
    if t == 0:
        x += addx
        line = next(lines)
        if line == "noop":
            # model it like an "addx 0" which takes only 1 cycle to complete
            addx, t = 0, 1
        else:
            _op, n = line.split()
            addx, t = int(n), 2
    t -= 1
    row, col = divmod(i, W)
    crt[row, col] = ".#"[abs(col - x) <= 1]


print("part a:", a)
print("part b:", AOCR[crt])



"""
--- Day 12: Digital Plumber ---
https://adventofcode.com/2017/day/12
"""
from aocd import data


def part_ab(data):
    d = {}
    for line in data.replace(" <->", ",").splitlines():
        xs = [int(x) for x in line.split(", ")]
        d[tuple(xs)] = set(xs)

    while True:
        for tuple_, set1 in d.items():
            try:
                match = next(k for k, set2 in d.items() if k != tuple_ and set1 & set2)
            except StopIteration:
                continue  # no match for this key - keep looking
            else:
                d[tuple_] = set1 | d.pop(match)
                break  # merged match and set1
        else:
            break  # no match for any key - we are done!

    [set0] = [v for v in d.values() if 0 in v]
    a = len(set0)
    b = len(d)
    return a, b


test_data = """\
0 <-> 2
1 <-> 1
2 <-> 0, 3, 4
3 <-> 2, 4
4 <-> 2, 3, 6
5 <-> 6
6 <-> 4, 5"""

assert part_ab(test_data) == (6, 2)

a, b = part_ab(data)
print("part a:", a)
print("part b:", b)



"""
--- Day 18: Like a Rogue ---
https://adventofcode.com/2016/day/18
"""
from aocd import data


def generate_rows(row):
    """rule90"""
    idx = range(len(row))
    while True:
        yield row
        row = "." + row + "."
        row = "".join(["^."[row[i] == row[i + 2]] for i in idx])


def make_grid(row0, nrows):
    gen = generate_rows(row0)
    return "".join([next(gen) + "\n" for i in range(nrows)])


def n_safe_tiles(row0, nrows):
    gen = generate_rows(row0)
    return sum([next(gen).count(".") for i in range(nrows)])


if __name__ == "__main__":
    print(n_safe_tiles(data, nrows=40))
    print(n_safe_tiles(data, nrows=400000))



"""
--- Day 8: Seven Segment Search ---
https://adventofcode.com/2021/day/8
"""
from aocd import data
from itertools import permutations

segments = {
    1: "cf",
    7: "acf",
    4: "bcdf",
    8: "abcdefg",

    2: "acdeg",
    3: "acdfg",
    5: "abdfg",

    6: "abdefg",
    9: "abcdfg",
    0: "abcefg",
}

lookup = {frozenset(v): str(k) for k, v in segments.items()}
translations = [str.maketrans("".join(p), "abcdefg") for p in permutations("abcdefg", 7)]

a = b = 0
for line in data.splitlines():
    patterns, code = line.split(" | ")
    for part in code.split():
        a += len(part) in {2, 3, 4, 7}
    for t in translations:
        if all(frozenset(w.translate(t)) in lookup for w in patterns.split()):
            break
    code = "".join([lookup[frozenset(n.translate(t))] for n in code.split()])
    b += int(code)

print("part a:", a)
print("part b:", b)



"""
--- Day 8: Two-Factor Authentication ---
https://adventofcode.com/2016/day/8
"""
import numpy as np
from aocd import data
from aoc_wim.ocr import AOCR


def animate(A, line):
    if line.startswith("rect"):
        w, h = [int(x) for x in line.split()[1].split("x")]
        A[0:h, 0:w] = "#"
    elif line.startswith("rotate"):
        i, shift = [int(x) for x in line.split("=")[1].split(" by ")]
        item = (i, slice(None)) if "row" in line else (slice(None), i)
        A[item] = np.roll(A[item], shift)
    return A


A = np.full((6, 50), ".")
for line in data.splitlines():
    A = animate(A, line)


if __name__ == "__main__":
    print("part a:", (A == "#").sum())
    print("part b:", AOCR[A])



"""
--- Day 6: Custom Customs ---
https://adventofcode.com/2020/day/6
"""
from aocd import data

a = b = 0
for block in data.split("\n\n"):
    s = {*block} - {"\n"}
    a += len(s)
    b += len(s.intersection(*block.splitlines()))

print(a)
print(b)



"""
--- Day 18: Like a GIF For Your Yard ---
https://adventofcode.com/2015/day/18
"""
import numpy as np
from aocd import data
from scipy.signal import convolve2d
from aoc_wim.zgrid import ZGrid


def parsed(data, part="a"):
    grid = ZGrid(data)
    A = (np.array(grid) == "#").astype(int)
    if part == "b":
        A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    return A


kernel = np.ones((3, 3), dtype=int)
kernel[1, 1] = 0


def evolve(A, part="a"):
    """conway's game of life"""
    C = convolve2d(A, kernel, mode="same")
    A = ((A == 1) & ((C == 2) | (C == 3))) | ((A == 0) & (C == 3))
    A = A.astype(int)
    if part == "b":
        A[0, 0] = A[0, -1] = A[-1, 0] = A[-1, -1] = 1
    return A


def animate(data, iterations=100, part="a"):
    A = parsed(data, part)
    for i in range(iterations):
        A = evolve(A, part)
    return A.sum()


if __name__ == "__main__":
    print(animate(data, part="a"))
    print(animate(data, part="b"))



"""
--- Day 8: Space Image Format ---
https://adventofcode.com/2019/day/8
"""
from aocd import data
import numpy as np
from aoc_wim.ocr import AOCR

rows, cols = 6, 25
a = np.fromiter(data, int).reshape((-1, rows, cols))
layer = min(a, key=lambda v: (v == 0).sum())
print((layer == 1).sum() * (layer == 2).sum())

img = np.ones_like(layer) * 2
for layer in a:
    np.copyto(img, layer, where=(img == 2))

txt = AOCR[img]
print(txt)



"""
--- Day 15: Rambunctious Recitation ---
https://adventofcode.com/2020/day/15
"""
from aocd import data


# CPython hack: loading names in a local namespace (LOAD_FAST op) is
# quicker than loading names in a module namespace (LOAD_GLOBAL op)
def local(n):
    seen = {}
    get = seen.get
    for i0, prev in enumerate(ns, start=1):
        seen[prev] = i0
    for i in range(i0, n):
        seen[prev], prev = i, i - get(prev, i)
    return prev


ns = [int(x) for x in data.split(",")]
print("part a:", local(2020))
print("part b:", local(30000000))



"""
--- Day 21: Chronal Conversion ---
https://adventofcode.com/2018/day/21
"""
from aocd import data
from parse import parse

template = """\
#ip {:d}
seti {:d} {:d} {:d}
bani {:d} {:d} {:d}
eqri {:d} {:d} {:d}
addr {:d} {:d} {:d}
seti {:d} {:d} {:d}
seti {:d} {:d} {:d}
bori {:d} {:d} {:d}
seti {bignum:d} {:d} {:d}
bani {:d} {:d} {:d}
addr {:d} {:d} {:d}
bani {:d} {:d} {:d}
muli {:d} {prime:d} {:d}
bani {:d} {:d} {:d}
gtir {:d} {:d} {:d}
addr {:d} {:d} {:d}
addi {:d} {:d} {:d}
seti {:d} {:d} {:d}
seti {:d} {:d} {:d}
addi {:d} {:d} {:d}
muli {:d} {:d} {:d}
gtrr {:d} {:d} {:d}
addr {:d} {:d} {:d}
addi {:d} {:d} {:d}
seti {:d} {:d} {:d}
addi {:d} {:d} {:d}
seti {:d} {:d} {:d}
setr {:d} {:d} {:d}
seti {:d} {:d} {:d}
eqrr {:d} {:d} {:d}
addr {:d} {:d} {:d}
seti {:d} {:d} {:d}"""
parsed = parse(template, data)


part_a = part_b = None
seen = set()
r3 = 0x10000
r5 = parsed.named["bignum"]
while True:
    r5 += r3 & 0xFF
    r5 &= 0xFFFFFF
    r5 *= parsed.named["prime"]
    r5 &= 0xFFFFFF
    if r3 >= 256:
        r3 //= 256
        continue
    if part_a is None:
        part_a = r5
    if r5 in seen:
        break
    part_b = r5
    seen.add(r5)
    r3 = r5 | 0x10000
    r5 = parsed.named["bignum"]

print(part_a)
print(part_b)



"""
--- Day 19: Monster Messages ---
https://adventofcode.com/2020/day/19
"""
from aocd import data
import regex as re


rules_raw, messages = data.replace('"', "").split("\n\n")
messages = messages.splitlines()
rules = dict(line.split(": ") for line in rules_raw.splitlines())
digits = set("0123456789")


def eliminate(rules, k0):
    v0 = rules.pop(k0)
    if "|" in v0:
        v0 = f"( {v0} )"
    for k, v in rules.items():
        rules[k] = " ".join([v0 if k == k0 else k for k in v.split()])
    return v0.replace(" ", "")


def simplify(rules):
    solved = {}
    unsolved = rules.copy()
    while True:
        for k, v in unsolved.items():
            if not digits.intersection(v):
                break
        else:
            break
        solved[k] = eliminate(unsolved, k)
    return solved, unsolved


solved, unsolved = simplify(rules)
r0a = re.compile("^" + solved["0"].replace(" ", "") + "$")
print("part a:", sum(r0a.match(m) is not None for m in messages))

if {"8", "11"} < rules.keys():
    rules["8"] += " | 42 8"
    rules["11"] += " | 42 11 31"
    solved, unsolved = simplify(rules)
    r42 = solved["42"].replace("(", "(?:")  # non-capture
    r31 = solved["31"].replace("(", "(?:")  # non-capture
    r8 = "(" + r42 + ")+"
    r11 = "(" + r42 + "(?2)?" + r31 + ")"
    r0b = re.compile("^" + r8 + r11 + "$")
    print("part b:", sum(r0b.match(m) is not None for m in messages))



"""
--- Day 18: Settlers of The North Pole ---
https://adventofcode.com/2018/day/18
"""
import time
import numpy as np
from aocd import data
from scipy.signal import convolve2d


def parsed(data):
    h = len(data.splitlines())
    A = np.fromiter(data.replace("\n", ""), dtype="<U1").reshape(h, -1)
    A = (A == "|").astype(complex) + (A == "#").astype(complex) * 1j
    return A


def draw(A, dt=0.1):
    # d = {0: ".", 1: "|", 1j: "#"}
    d = {0: "  ", 1: "🌲", 1j: " ⛏"}
    print("\33c")
    for row in A:
        print(*[d[c] for c in row], sep="")
    print()
    time.sleep(dt)


def mutate(A0):
    k = np.ones((3, 3), dtype=complex)
    B = convolve2d(A0, k, mode="same")
    A1 = A0.copy()
    A1[(A0 == 0) & (B.real >= 3)] = 1
    A1[(A0 == 1) & (B.imag >= 3)] = 1j
    A1[(A0 == 1j) & ((B.imag < 2) | (B.real == 0))] = 0
    return A1


def evolve(data, minutes=10):
    A = parsed(data)
    m = minutes
    seen = {A.tobytes(): m}
    while m > 0:
        m -= 1
        A = mutate(A)
        # draw(A)
        s = A.tobytes()
        if s in seen:
            delta = seen[s] - m
            m %= delta
        seen[s] = m
    return (A == 1).sum() * (A == 1j).sum()


print("part a:", evolve(data, minutes=10))
print("part b:", evolve(data, minutes=1000000000))



"""
--- Day 12: Hill Climbing Algorithm ---
https://adventofcode.com/2022/day/12
"""
from aocd import data
from aoc_wim.zgrid import ZGrid
from aoc_wim.search import AStar

grid = ZGrid(data)
start = grid.z("S")
end = grid.z("E")
grid[start] = "a"
grid[end] = "z"
starts = grid.z("a", first=False)


class Q12AStar(AStar):
    def adjacent(self, z0):
        return [z for z in grid.near(z0) if ord(grid.get(z, "⛰")) <= ord(grid[z0]) + 1]


a = len(Q12AStar(start, end).run()) - 1
b = min(len(path) - 1 for s in starts if (path := Q12AStar(s, end).run()) is not None)
print("part a:", a)
print("part b:", b)



"""
--- Day 25: Combo Breaker ---
https://adventofcode.com/2020/day/25
"""
from aocd import data


def crack(pk_card, pk_door, subject_number=7, modulus=20201227):
    loop_size = 0
    p = 1
    while True:
        loop_size += 1
        p = p * subject_number % modulus
        if p == pk_card:
            return pow(pk_door, loop_size, modulus)
        if p == pk_door:
            return pow(pk_card, loop_size, modulus)


pk_card, pk_door = data.split()
print(crack(int(pk_card), int(pk_door)))



"""
--- Day 4: Secure Container ---
https://adventofcode.com/2019/day/4
"""
from aocd import data


def ok(n, part="a"):
    # It is a six-digit number.
    if not (100_000 <= n <= 999_999):
        return False

    # Two adjacent digits are the same. (part a)
    # ... but not part of a larger group of matching digits (part b)
    str_n = str(n)
    for s in "0123456789":
        if 2 * s in str_n:
            if part == "a":
                break
            elif part == "b" and 3 * s not in str_n:
                break
    else:
        return False

    # Going from left to right, the digits never decrease
    for i in range(5):
        if str_n[i + 1] < str_n[i]:
            return False

    return True


lo, hi = data.split("-")
ns = range(int(lo), int(hi) + 1)

print("part a:", sum(ok(n, part="a") for n in ns))
print("part b:", sum(ok(n, part="b") for n in ns))



"""
--- Day 9: Sensor Boost ---
https://adventofcode.com/2019/day/9
"""
from aocd import data
from aoc_wim.aoc2019 import IntComputer


def compute(data, inputs=()):
    comp = IntComputer(data, inputs=inputs)
    comp.run()
    result = ",".join([str(x) for x in reversed(comp.output)])
    return result


if __name__ == "__main__":
    print(compute(data, inputs=[1]))
    print(compute(data, inputs=[2]))



"""
--- Day 7: Recursive Circus ---
https://adventofcode.com/2017/day/7
"""
from collections import Counter
from collections import deque

from aocd import data
from wimpy import cached_property


class Node:
    def __init__(self, name):
        self.name = name
        self.weight = None
        self.parent = None
        self.children = []

    @cached_property
    def rweight(self):
        return self.weight + sum(x.rweight for x in self.children)

    @property
    def siblings(self):
        if self.parent is None:
            return []
        return [x for x in self.parent.children if x is not self]

    @classmethod
    def get_or_create(cls, name):
        if name not in nodes:
            nodes[name] = cls(name)
        return nodes[name]


def make_tree(data):
    for line in data.splitlines():
        line = line.translate(str.maketrans("", "", "()->,"))
        name, weight, *child_names = line.split()
        node = Node.get_or_create(name)
        node.weight = int(weight)
        for name in child_names:
            child = Node.get_or_create(name)
            child.parent = node
            node.children.append(child)
    # find root
    while node.parent:
        node = node.parent
    return node


def find_bad_node_correct_weight(tree):
    d = deque([tree])
    while d:
        node = d.pop()
        c = Counter(x.rweight for x in node.children)
        for child in node.children:
            if c[child.rweight] == 1:
                d.append(child)
                break
    delta = node.rweight - node.siblings[0].rweight
    return node.weight - delta


nodes = {}
tree = make_tree(data)
print(tree.name)
print(find_bad_node_correct_weight(tree))



"""
--- Day 20: Donut Maze ---
https://adventofcode.com/2019/day/20
"""
from aocd import data
from aoc_wim.zgrid import ZGrid
from aoc_wim.search import AStar
from bidict import bidict
import string
import numpy as np

grid = ZGrid(data, on=".", off="#")
h, w = np.array(grid).shape
dzs = [-1j, 1, 1j, -1]

# parse the warps
outside = {}
inside = {}
for z, glyph in grid.items():
    if glyph in string.ascii_uppercase:
        for dz in dzs:
            if grid.get(z + dz) == ".":
                zp = z + dz  # actual position of portal
                name = glyph + grid.get(z - dz)  # add other letter
                if 3 < z.real < w - 3 and 3 < z.imag < h - 3:
                    side = inside
                else:
                    side = outside
                if zp - z in {1j, 1}:
                    name = name[::-1]  # reverse the label
                side[name] = zp
                break

state0 = outside.pop("AA")
target = outside.pop("ZZ")
assert outside.keys() == inside.keys()
warps = bidict({v: outside[k] for k, v in inside.items()})


class Q20AStar(AStar):

    def __init__(self, part="a", maxdepth=30):
        self.d_level = 1 if part == "b" else 0
        AStar.__init__(self, (state0, 0), (target, 0))
        self.maxdepth = maxdepth

    def adjacent(self, state):
        z0, level = state
        if level > self.maxdepth:
            return
        if z0 in warps:
            yield warps[z0], level + self.d_level
        if z0 in warps.inv:
            if level > 0 or self.d_level == 0:
                yield warps.inv[z0], level - self.d_level
        for dz in dzs:
            z = z0 + dz
            if grid.get(z) == ".":
                yield z, level


astar = Q20AStar(part="a")
astar.run()
print(astar.path_length)

astar = Q20AStar(part="b")
astar.run()
print(astar.path_length)



import logging


log = logging.getLogger(__name__)


class Comp:

    def __init__(self, ops):
        self.ops = ops  # parsed code
        self.a = 0      # accumulator
        self.line = 0   # instruction pointer into ops
        self.n = 0      # number of instructions exec

    def acc(self, a):
        self.a += a

    def jmp(self, i):
        self.line += i - 1

    def nop(self, _):
        pass

    def step(self):
        self.n += 1
        op, arg = self.ops[self.line]
        op_func = getattr(self, op)
        log.debug(f"{self.n:5}:   {op:3}{arg:4}   L={self.line:3}   a={self.a:4}")
        op_func(arg)
        self.line += 1


def find_pair(counter, target=2020):
    # find a pair of numbers from the multiset (counter) which sums to target
    # if the input is a set, find a pair of *distinct* numbers
    for number in counter:
        diff = target - number
        if diff in counter:
            if diff == number and (isinstance(counter, set) or counter[number] <= 1):
                continue
            return number, diff



"""
--- Day 23: Experimental Emergency Teleportation ---
https://adventofcode.com/2018/day/23
"""
import z3
from aocd import get_data
from parse import parse


def z3_abs(x):
    return z3.If(x >= 0, x, -x)


def z3_dist(x, y):
    return z3_abs(x[0] - y[0]) + z3_abs(x[1] - y[1]) + z3_abs(x[2] - y[2])


def dist(x, y):
    return abs(x[0] - y[0]) + abs(x[1] - y[1]) + abs(x[2] - y[2])


data = get_data(year=2018, day=23)
template = "pos=<{:d},{:d},{:d}>, r={:d}"
bots = [parse(template, s).fixed for s in data.splitlines()]
x = z3.Int("x")
y = z3.Int("y")
z = z3.Int("z")
orig = (x, y, z)
cost_expr = x * 0
for *pos, r in bots:
    cost_expr += z3.If(z3_dist(orig, pos) <= r, 1, 0)
opt = z3.Optimize()
opt.maximize(cost_expr)
opt.minimize(z3_dist((0, 0, 0), (x, y, z)))
opt.check()
model = opt.model()
v = [model[c].as_long() for c in (x, y, z)]
print("best location:", v)
print("bots in range:", sum(1 for bot in bots if dist(bot[:3], v) < bot[3]))
print("distance from origin:", dist((0, 0, 0), v))



"""
--- Day 13: Mine Cart Madness ---
https://adventofcode.com/2018/day/13
"""
import random
from collections import Counter
from itertools import cycle

from aocd import data
from bidict import bidict
from termcolor import colored
from termcolor import COLORS


class CartCollision(Exception):
    def __init__(self, coordinates):
        self.coordinates = coordinates
        super().__init__(coordinates)


class Cart:

    vs = bidict(zip("^>v<", [-1j, 1, 1j, -1]))

    def __init__(self, x, y, glyph, color=None):
        self.glyph = glyph
        self._position = complex(x, y)
        self._turns = cycle([-1j, 1, 1j])  # left, straight, right
        self.color = color

    def __repr__(self):
        return f"<Cart {self.pretty_glyph} at {self.coordinates}>"

    def __lt__(self, other):
        if not isinstance(other, Cart):
            return NotImplemented
        return (self.y, self.x) < (other.y, other.x)

    @property
    def pretty_glyph(self):
        return colored(self.glyph, self.color, attrs=["bold"])

    @property
    def velocity(self):
        return self.vs[self.glyph]

    @property
    def y(self):
        # row
        return int(self._position.imag)

    @property
    def x(self):
        # column
        return int(self._position.real)

    @property
    def coordinates(self):
        return f"{self.x},{self.y}"

    def turn(self, direction=None):
        if direction is None:
            direction = next(self._turns)
        v = self.velocity * direction
        self.glyph = self.vs.inv[v]

    def tick(self, grid):
        # draw(grid, carts=[self])
        self._position += self.velocity
        track = grid[self.y, self.x]
        assert track in r"+/-\|", f"WTF: {self} ran off the rails"
        if track == "+":
            direction = None
        elif track == "\\":
            direction = -1j if self.glyph in "v^" else 1j
        elif track == "/":
            direction = -1j if self.glyph in "><" else 1j
        else:
            return
        self.turn(direction)


def parsed(data):
    grid = {}
    carts = []
    lines = data.splitlines()
    h = len(lines)
    w = max(len(line) for line in lines)
    for y, line in enumerate(lines):
        for x, char in enumerate(line):
            if char in "v>^<":
                grid[y, x] = "|" if char in "v^" else "-"
                cart = Cart(x, y, glyph=char, color=random.choice(list(COLORS)))
                carts.append(cart)
            elif char != " ":
                grid[y, x] = char
    grid["y-axis"] = range(h)
    grid["x-axis"] = range(w)
    return grid, carts


def draw(grid, carts):
    seen = set()
    carts_on_grid = {}
    prompt = ""
    for cart in carts:
        if cart.coordinates in seen:
            # collision
            prompt = "Collision detected! "
            val = colored("X", "red", attrs=["bold"])
        else:
            val = cart.pretty_glyph
        carts_on_grid[cart.y, cart.x] = val
        seen.add(cart.coordinates)
    print()
    for y in grid["y-axis"]:
        line = []
        for x in grid["x-axis"]:
            pos = (y, x)
            if pos in carts_on_grid:
                line.append(carts_on_grid[pos])
            else:
                line.append(grid.get(pos, " "))
        print(*line, sep="")
    prompt += "Press enter to continue..."
    input(prompt)
    print("\33c")


def find_collision(carts):
    counter = Counter([c.coordinates for c in carts])
    pos = max(counter, key=counter.get)
    if counter[pos] > 1:
        raise CartCollision(pos)


def part_a(data, first_crash=True):
    a0, carts = parsed(data)
    while True:
        # draw(a0, carts)
        carts.sort()
        for cart in carts:
            cart.tick(grid=a0)
            try:
                find_collision(carts)
            except CartCollision as err:
                # draw(a0, carts)
                if first_crash:
                    return err.coordinates
                # remove exactly 2 crashed carts. there can not be more than
                # one collision at a time because we only moved 1 cart at a
                # time. this is not strictly correct approach because 3 or 4
                # carts could theoretically crash together at an intersection
                carts = [c for c in carts if c.coordinates != err.coordinates]
        if not carts:
            return
        if len(carts) == 1:
            [cart] = carts
            return cart.coordinates


def part_b(data):
    return part_a(data, first_crash=False)


print("part a:", part_a(data))
print("part b:", part_b(data))



"""
--- Day 23: Safe Cracking ---
https://adventofcode.com/2016/day/23
"""
from aocd import data
from aoc_wim.aoc2016 import AssembunnyComputer


comp = AssembunnyComputer(data, a0=7)
comp.run()
print("part a:", comp.reg["a"])

comp.__init__(data, a0=12)
comp.run()
print("part b:", comp.reg["a"])



"""
--- Day 14: Reindeer Olympics ---
https://adventofcode.com/2015/day/14
"""
from aocd import data


class Reindeer(object):
    def __init__(self, line):
        words = line.split()
        self.name = words[0]
        self.run_speed = int(words[3])
        self.stamina = int(words[6])
        self.rest_time = int(words[-2])
        self.distance = 0
        self.points = 0
        self.speed = self.run_speed
        self.dt = self.stamina

    def step(self):
        self.distance += self.speed
        self.dt -= 1
        if self.dt == 0:
            if self.speed == 0:
                self.speed = self.run_speed
                self.dt = self.stamina
            else:
                self.speed = 0
                self.dt = self.rest_time


def race(data, max_t, measure):
    deers = [Reindeer(line) for line in data.splitlines()]
    for t in range(max_t):
        for deer in deers:
            deer.step()
        max_distance = max([deer.distance for deer in deers])
        for deer in deers:
            if deer.distance == max_distance:
                deer.points += 1
    return max([getattr(deer, measure) for deer in deers])


if __name__ == "__main__":
    print(race(data, max_t=2503, measure="distance"))
    print(race(data, max_t=2503, measure="points"))



"""
--- Day 20: Trench Map ---
https://adventofcode.com/2021/day/20
"""
from aocd import data
import numpy as np
from scipy.signal import convolve2d

s, grid = data.split("\n\n")
s = np.array([c == "#" for c in s])
im = np.array([[*r] for r in grid.splitlines()]) == "#"
kernel = 2 ** np.arange(9).reshape(3, 3)
for i in range(50):
    im = s[convolve2d(im, kernel, fillvalue=s[0]*i % 2)]
    if i == 1:
        print("part a:", im.sum())
print("part b:", im.sum())



"""
--- Day 9: Explosives in Cyberspace ---
https://adventofcode.com/2016/day/9
"""
from collections import deque
import re
import logging

from aocd import data


log = logging.getLogger(__name__)


pat = re.compile(r"\((?P<n_chars>[0-9]+)x(?P<n_repeats>[0-9]+)\)")


def safe_to_use_fast_algo(data):
    # checks if all assumptions about the data actually hold before using faster path
    # does every "(" start a valid marker?
    for i, char in enumerate(data):
        if char == "(":
            match = pat.match(data, i)
            if match is None:
                log.debug("%d: '(' didn't start a valid marker", i)
                return False
            n_chars = int(match.group("n_chars"))
            # the substring to be repeated must not exceed the remaining length of data
            if n_chars > len(data) - match.end():
                log.debug("%d: substring to be repeated exceeded end", i)
                return False
            for j in range(match.end(), match.end() + n_chars):
                match_nested = pat.match(data, j)
                if match_nested is None:
                    continue
                assert match_nested.start() < match.end() + n_chars
                assert match_nested.start() == j
                # don't split up other markers when repeating
                if match.end() + n_chars < match_nested.end():
                    log.debug("%d: substring to repeated split another marker", i)
                    return False
                n_chars_nested = int(match_nested.group("n_chars"))
                # Can't copy an inner marker that wants to access data
                # outside of the outer marker's reach
                if match.end() + n_chars < match_nested.end() + n_chars_nested:
                    log.debug("%d: inner marker reaches beyond outer marker range", i)
                    return False
    return True


def parse_marker(d, i=0):
    if len(d) - i < 5:
        # need at least 5 chars to make a valid marker (MxN)
        return
    if d[i] != "(":
        # marker must start with (
        return
    vals = []
    for val in d:
        vals.append(val)
        if val == ")":
            break
    assert vals[0] == "("
    if vals[-1] != ")":
        # marker must end with )
        return
    substring = "".join(vals[1:-1])
    try:
        duration, multiplier = [int(x) for x in substring.split("x")]
    except ValueError:
        # in between ( and ) must be number x number
        return
    width = len(substring) + 2
    assert width >= 5
    return duration, multiplier, width


def unzip_slow(data, part="a"):
    length = 0
    d = deque(data)
    while d:
        marker = parse_marker(d)
        if marker is None:
            d.popleft()
            length += 1
            continue
        duration, multiplier, width = marker
        for _ in range(width):
            d.popleft()
        if part == "a":
            for _ in range(min(duration, len(d))):
                d.popleft()
                length += multiplier
        elif part == "b":
            extend = [d[i] for i in range(min(duration, len(d)))] * (multiplier - 1)
            d.extendleft(reversed(extend))
    return length


def unzip_fast(data, part="a"):
    length = 0
    i = 0
    while True:
        m = pat.search(data, i)
        if m is None:
            length += len(data) - i
            break
        n = int(m.group("n_chars"))
        r = int(m.group("n_repeats"))
        if part == "a":
            length += m.start() - i + n * r
        elif part == "b":
            length += m.start() - i + unzip_fast(data[m.end() : m.end() + n], part) * r
        i = m.end() + n
    return length


if safe_to_use_fast_algo(data):
    print("part a:", unzip_fast(data, part="a"))
    print("part b:", unzip_fast(data, part="b"))
else:
    print("part a:", unzip_slow(data, part="a"))
    print("part b:", unzip_slow(data, part="b"))



"""
--- Day 16: Aunt Sue ---
https://adventofcode.com/2015/day/16
"""
from collections import defaultdict
from operator import eq
from operator import gt
from operator import lt

from aocd import data


message = """\
children: 3
cats: 7
samoyeds: 2
pomeranians: 3
akitas: 0
vizslas: 0
goldfish: 5
trees: 3
cars: 2
perfumes: 1
"""


d_message = {k: int(v) for k, v in (pair.split(": ") for pair in message.splitlines())}

sues = {}
for line in data.splitlines():
    sue, stats = line.split(": ", 1)
    stats = {k: int(v) for k, v in (pair.split(": ") for pair in stats.split(", "))}
    sues[sue] = stats


def find_sue(**kwargs):
    opmap = defaultdict(lambda: eq)
    opmap.update(kwargs)
    for sue, stats in sues.items():
        common_keys = d_message.keys() & stats.keys()
        if all(opmap[k](stats[k], d_message[k]) for k in common_keys):
            return int(sue.split()[1])


print(find_sue())
print(find_sue(cats=gt, trees=gt, pomeranians=lt, goldfish=lt))
